@@ -8,13 +8,14 @@ import numpy as np
 from pandas import DataFrame
 from scipy.optimize import curve_fit, minimize, OptimizeWarning
 from scipy.spatial.distance import pdist, squareform
-from scipy import stats 
+from scipy import stats, sparse
 from sklearn.isotonic import IsotonicRegression
 
 from skgstat import estimators, models, binning
 from skgstat import plotting
 from skgstat.util import shannon_entropy
-
+from .MetricSpace import MetricSpace, MetricSpacePair
+from skgstat.interfaces.gstools import skgstat_to_gstools
 
 class Variogram(object):
     """Variogram Class
@@ -42,18 +43,24 @@ class Variogram(object):
                  ):
         r"""Variogram Class
 
-        Note: The directional variogram estimation is not re-implemented yet.
-        Therefore the parameters is-directional, azimuth and tolerance will
-        be ignored at the moment and can be subject to changes.
-
         Parameters
         ----------
-        coordinates : numpy.ndarray
+        coordinates : numpy.ndarray, MetricSpace
+            .. versionchanged:: 0.5.0
+                now accepts MetricSpace
             Array of shape (m, n). Will be used as m observation points of
             n-dimensions. This variogram can be calculated on 1 - n
             dimensional coordinates. In case a 1-dimensional array is passed,
             a second array of same length containing only zeros will be
             stacked to the passed one.
+            For very large datasets, you can set maxlag to only calculate
+            distances within the maximum lag in a sparse matrix.
+            Alternatively you can supply a MetricSpace (optionally with a
+            `max_dist` set for the same effect). This is useful if you're
+            creating many different variograms for different measured
+            parameters that are all measured at the same set of coordinates,
+            as distances will only be calculated once, instead of once per
+            variogram.
         values : numpy.ndarray
             Array of values observed at the given coordinates. The length of
             the values array has to match the m dimension of the coordinates
@@ -92,7 +99,7 @@ class Variogram(object):
             scipy.spatial.distance.pdist. Additional parameters are not (yet)
             passed through to pdist. These are accepted by pdist for some of
             the metrics. In these cases the default values are used.
-        bin_func : str            
+        bin_func : str
             .. versionchanged:: 0.3.8
                 added 'fd', 'sturges', 'scott', 'sqrt', 'doane'
             .. versionchanged:: 0.3.9
@@ -190,7 +197,7 @@ class Variogram(object):
         percentile : int
             .. versionadded:: 0.3.7
 
-            If the `estimator <skgstat.Variogram.estimator>` is set to 
+            If the `estimator <skgstat.Variogram.estimator>` is set to
             `'entropy'` this argument sets the percentile to be used.
         binning_random_state : int, None
             .. versionadded:: 0.3.9
@@ -211,8 +218,27 @@ class Variogram(object):
         # Before we do anything else, make kwargs available
         self._kwargs = self._validate_kwargs(**kwargs)
 
+        # handle the coordinates
+        self._1d = False
+        if not isinstance(coordinates, MetricSpace):
+            coordinates = np.asarray(coordinates)
+
+            # handle 1D coords
+            if len(coordinates.shape) < 2:
+                coordinates = np.column_stack((
+                    coordinates,
+                    np.zeros(len(coordinates))
+                ))
+                self._1d = True
+
+            # handle maxlag for MetricSpace
+            _maxlag = maxlag if maxlag and not isinstance(maxlag, str) and maxlag >= 1 else None
+            coordinates = MetricSpace(coordinates.copy(), dist_func, _maxlag)
+        elif dist_func != coordinates.dist_metric:
+            raise AttributeError("Distance metric of variogram differs from distance metric of coordinates")
+
         # Set coordinates
-        self._X = np.asarray(coordinates)
+        self._X = coordinates
 
         # pairwise differences
         self._diff = None
@@ -225,13 +251,6 @@ class Variogram(object):
         # calc_diff = False here, because it will be calculated by fit() later
         self.set_values(values=values, calc_diff=False)
 
-        # distance matrix
-        self._dist = None
-
-        # set distance calculation function
-        self._dist_func_name = None
-        self.set_dist_function(func=dist_func)
-
         # lags and max lag
         self._n_lags_passed_value = n_lags
         self._n_lags = None
@@ -246,10 +265,6 @@ class Variogram(object):
         self._estimator = None
         self.set_estimator(estimator_name=estimator)
 
-        # model can be a function or a string
-        self._model = None
-        self.set_model(model_name=model)
-
         # the binning settings
         self._bin_func_name = None
         self._bin_func = None
@@ -257,6 +272,13 @@ class Variogram(object):
         self._bins = None
         self.set_bin_func(bin_func=bin_func)
 
+        # Needed for harmonized models
+        self.preprocessing(force=True)
+
+        # model can be a function or a string
+        self._model = None
+        self.set_model(model_name=model)
+
         # specify if the lag should be given absolute or relative to the maxlag
         self._normalized = normalize
 
@@ -293,7 +315,16 @@ class Variogram(object):
         coordinates : numpy.array
 
         """
-        return self._X
+        return self._X.coords
+
+    @property
+    def dim(self):
+        """
+        Input coordinates dimensionality.
+        """
+        if self._1d:
+            return 1
+        return self.coordinates.shape[1]
 
     @property
     def values(self):
@@ -369,13 +400,13 @@ class Variogram(object):
 
         """
         # check dimensions
-        if not len(values) == len(self._X):
+        if not len(values) == len(self.coordinates):  # pragma: no cover
             raise ValueError('The length of the values array has to match' +
                              'the length of coordinates')
 
         # use an array
         _y = np.asarray(values)
-        if not _y.ndim == 1:
+        if not _y.ndim == 1:  # pragma: no cover
             raise ValueError('The values shall be a 1-D array.' +
                              'Multi-dimensional values not supported yet.')
 
@@ -537,33 +568,36 @@ class Variogram(object):
 
         References
         ----------
-        .. [101] Scott, D.W. (2009), Sturges' rule. WIREs Comp Stat, 1: 303-306. 
+        .. [101] Scott, D.W. (2009), Sturges' rule. WIREs Comp Stat, 1: 303-306.
             https://doi.org/10.1002/wics.35
-        .. [102] Scott, D.W. (2010), Scott's rule. WIREs Comp Stat, 2: 497-502. 
+        .. [102] Scott, D.W. (2010), Scott's rule. WIREs Comp Stat, 2: 497-502.
             https://doi.org/10.1002/wics.103
-        .. [103] Freedman, David, and Persi Diaconis  (1981), "On the histogram as 
-            a density estimator: L 2 theory." Zeitschrift für Wahrscheinlichkeitstheorie 
+        .. [103] Freedman, David, and Persi Diaconis  (1981), "On the histogram as
+            a density estimator: L 2 theory." Zeitschrift für Wahrscheinlichkeitstheorie
             und verwandte Gebiete 57.4: 453-476.
-        .. [104] Doane, D. P. (1976). Aesthetic frequency classifications. 
+        .. [104] Doane, D. P. (1976). Aesthetic frequency classifications.
             The American Statistician, 30(4), 181-183.
 
         """
-        # switch the input
-        if bin_func.lower() == 'even':
-            self._bin_func = binning.even_width_lags
+        # handle strings
+        if isinstance(bin_func, str):
+            # switch the input
+            if bin_func.lower() == 'even':
+                self._bin_func = binning.even_width_lags
 
-        elif bin_func.lower() == 'uniform':
-            self._bin_func = binning.uniform_count_lags
+            elif bin_func.lower() == 'uniform':
+                self._bin_func = binning.uniform_count_lags
 
-        elif isinstance(bin_func, str):
             # remove the n_lags if they will be adjusted on call
-            if bin_func.lower() not in ('kmeans', 'ward', 'stable_entropy'):
-                self._n_lags = None
+            else:
+                # reset lags for adjusting algorithms
+                if bin_func.lower() not in ('kmeans', 'ward', 'stable_entropy'):
+                    self._n_lags = None
 
-            # use the wrapper
-            self._bin_func = self._bin_func_wrapper
+                # use the wrapper for all but even and uniform
+                self._bin_func = self._bin_func_wrapper
 
-        elif callable(bin_func):
+        elif callable(bin_func):  # pragma: no cover
             self._bin_func = bin_func
             bin_func = 'custom'
 
@@ -581,7 +615,7 @@ class Variogram(object):
     def _bin_func_wrapper(self, distances, n, maxlag):
         """
         Wrapper arounf the call of the actual binning method.
-        This is needed to pass keyword arguments to kmeans or 
+        This is needed to pass keyword arguments to kmeans or
         stable_entropy binning methods, and respect the slightly
         different function signature of auto_derived_lags.
         """
@@ -662,7 +696,7 @@ class Variogram(object):
     def n_lags(self, n):
         # TODO: here accept strings and implement some optimum methods
         # string are not implemented yet
-        if isinstance(n, str):
+        if isinstance(n, str):  # pragma: no cover
             raise NotImplementedError('n_lags string values not implemented')
 
         # n_lags is int
@@ -724,7 +758,7 @@ class Variogram(object):
                         'provide the function.'
                     ) % estimator_name
                 )
-        elif callable(estimator_name):
+        elif callable(estimator_name):  # pragma: no cover
             self._estimator = estimator_name
         else:
             raise ValueError('The estimator has to be a string or callable.')
@@ -748,21 +782,11 @@ class Variogram(object):
         if isinstance(model_name, str):
             # at first reset harmonize
             self._harmonize = False
-            if model_name.lower() == 'spherical':
-                self._model = models.spherical
-            elif model_name.lower() == 'exponential':
-                self._model = models.exponential
-            elif model_name.lower() == 'gaussian':
-                self._model = models.gaussian
-            elif model_name.lower() == 'cubic':
-                self._model = models.cubic
-            elif model_name.lower() == 'stable':
-                self._model = models.stable
-            elif model_name.lower() == 'matern':
-                self._model = models.matern
-            elif model_name.lower() == 'harmonize':
+            if model_name.lower() == 'harmonize':
                 self._harmonize = True
                 self._model = self._build_harmonized_model()
+            elif hasattr(models, model_name.lower()):
+                self._model = getattr(models, model_name.lower())
             else:
                 raise ValueError(
                     (
@@ -770,7 +794,7 @@ class Variogram(object):
                         ' understood, please provide the function'
                     ) % model_name
                 )
-        else:
+        else:  # pragma: no cover
             self._model = model_name
 
     def _build_harmonized_model(self):
@@ -786,23 +810,27 @@ class Variogram(object):
             """Monotonized Variogram
 
             Return the isotonic harmonized experimental variogram.
-            This means, the experimental variogram is monotonic after harmonization.
+            This means, the experimental variogram is monotonic
+            after harmonization.
 
-            The harmonization is done using following Hinterding (2003) using 
-            the PAVA algorithm (Barlow and Bartholomew, 1972).
+            The harmonization is done using following Hinterding (2003)
+            using the PAVA algorithm (Barlow and Bartholomew, 1972).
 
             Returns
             -------
             gamma : numpy.ndarray
                 monotonized experimental variogram
-            
+
             References
             ----------
-            Barlow, R., D. Bartholomew, et al. (1972): Statistical Interference Under Order Restrictions.
-                John Wiley and Sons, New York.
-            Hiterding, A. (2003): Entwicklung hybrider Interpolationsverfahren für den automatisierten Betrieb am
-                Beispiel meteorologischer Größen. Dissertation, Institut für Geoinformatik, Westphälische
-                Wilhelms-Universität Münster, IfGIprints, Münster. ISBN: 3-936616-12-4
+            Barlow, R., D. Bartholomew, et al. (1972): Statistical
+                Interference Under Order Restrictions. John Wiley
+                and Sons, New York.
+            Hiterding, A. (2003): Entwicklung hybrider Interpolations-
+                verfahren für den automatisierten Betrieb am Beispiel
+                meteorologischer Größen. Dissertation, Institut für
+                Geoinformatik, Westphälische Wilhelms-Universität Münster,
+                IfGIprints, Münster. ISBN: 3-936616-12-4
 
             """
 
@@ -827,13 +855,14 @@ class Variogram(object):
 
     @property
     def dist_function(self):
-        return self._dist_func_name
+        return self._X.dist_metric
 
-    def _dist_func_wrapper(self, x):
-        if callable(self._dist_func_name):
-            return self._dist_func_name(x)
+    @classmethod
+    def wrapped_distance_function(cls, dist_func, x):
+        if callable(dist_func):
+            return dist_func(x)
         else:
-            return pdist(X=x, metric=self._dist_func_name)
+            return pdist(X=x, metric=dist_func)
 
     @dist_function.setter
     def dist_function(self, func):
@@ -858,38 +887,46 @@ class Variogram(object):
         numpy.array
 
         """
-        # reset the distances and fitting
-        self._dist = None
+        # reset the fitting
         self.cof, self.cov = None, None
 
-        if isinstance(func, str):
+        if isinstance(func, str):  # pragma: no cover
             if func.lower() == 'rank':
                 raise NotImplementedError
-            else:
-                # if not ranks, it has to be a scipy metric
-                self._dist_func_name = func
-
-        elif callable(func):
-            self._dist_func_name = func
-        else:
+        elif not callable(func):
             raise ValueError('Input not supported. Pass a string or callable.')
 
         # re-calculate distances
-        self._calc_distances()
+        self._X = MetricSpace(self._X.coords, func, self._X.max_dist)
 
     @property
     def distance(self):
-        if self._dist is None:
-            self._calc_distances()
-        return self._dist
+        # handle sparse matrix
+        if isinstance(self.distance_matrix, sparse.spmatrix):
+            return self.triangular_distance_matrix.data
+        
+        # Turn it back to triangular form not to have duplicates
+        return squareform(self.distance_matrix)
 
-    @distance.setter
-    def distance(self, dist_array):
-        self._dist = dist_array
+    @property
+    def triangular_distance_matrix(self):
+        """
+        Like distance_matrix but with zeros below the diagonal...
+        Only defined if distance_matrix is a sparse matrix
+        """
+        if not isinstance(self.distance_matrix, sparse.spmatrix):
+            raise RuntimeWarning("Only available for sparse coordinates.")
+
+        m = self.distance_matrix
+        c = m.tocsc()
+        c.data = c.indices
+        rows = c.tocsr()
+        filt = sparse.csr.csr_matrix((m.indices < rows.data, m.indices, m.indptr))
+        return m.multiply(filt)
 
     @property
     def distance_matrix(self):
-        return squareform(self.distance)
+        return self._X.dists
 
     @property
     def maxlag(self):
@@ -903,7 +940,7 @@ class Variogram(object):
         # remove bins
         self._bins = None
         self._groups = None
-
+        
         # set new maxlag
         if value is None:
             self._maxlag = None
@@ -926,7 +963,7 @@ class Variogram(object):
         cost function, which divides the residuals by their uncertainty.
 
         When setting fit_sigma, the array of uncertainties itself can be
-        given, or one of the strings: ['linear', 'exp', 'sqrt', 'sq', 'entropy']. 
+        given, or one of the strings: ['linear', 'exp', 'sqrt', 'sq', 'entropy'].
         The parameters described below refer to the setter of this property.
 
         .. versionchanged:: 0.3.11
@@ -949,7 +986,7 @@ class Variogram(object):
                 :math:`w = \sqrt(w_n)`
               * **sigma='sq'**: The residuals get weighted by the function:
                 :math:`w = w_n^2`
-              * **sigma='entropy'**: Calculates the Shannon Entropy as 
+              * **sigma='entropy'**: Calculates the Shannon Entropy as
                 intrinsic uncertainty of each lag class.
 
         Returns
@@ -1036,9 +1073,9 @@ class Variogram(object):
         need to be updated.
 
         .. note::
-            Updating the kwargs does not force a preprocessing circle. 
-            Any affected intermediate result, that might be cached internally, 
-            will not make use of updated kwargs. Make a call to 
+            Updating the kwargs does not force a preprocessing circle.
+            Any affected intermediate result, that might be cached internally,
+            will not make use of updated kwargs. Make a call to
             :func:`preprocessing(force=True) <skgstat.Variogram.preprocessing>`
             to force a clean re-calculation of the Variogram instance.
 
@@ -1123,7 +1160,6 @@ class Variogram(object):
 
         """
         # call the _calc functions
-        self._calc_distances(force=force)
         self._calc_diff(force=force)
         self._calc_groups(force=force)
 
@@ -1370,33 +1406,39 @@ class Variogram(object):
         # get the pars
         cof = self.cof
 
-        # get the function
-        func = self._model
+        return self.fitted_model_function(self._model, self.cof)
 
-        if self._harmonize:
-            code = """model = lambda x: func(x)"""
+    @classmethod
+    def fitted_model_function(cls, model, cof=None, **kw):
+        if cof is None:
+            # Make sure to keep this synchronized with the output
+            # of describe()!
+            cof = [kw["effective_range"], kw["sill"]]
+            if "smoothness" in kw:
+                cof.append(kw["smoothness"])
+            if "shape" in kw:
+                cof.append(kw["shape"])
+            if kw["nugget"] != 0:
+                cof.append(kw["nugget"])
+
+        harmonize = False
+        if not callable(model):
+            if model == "harmonize":
+                raise ValueError("Please supply the actual harmonized model directly")
+            else:
+                model = model.lower()
+                model = getattr(models, model)
+            
+        if model.__name__ == "harmonize":
+            code = """fitted_model = lambda x: model(x)"""
         else:
-            code = """model = lambda x: func(x, %s)""" % \
+            code = """fitted_model = lambda x: model(x, %s)""" % \
                (', '.join([str(_) for _ in cof]))
 
         # run the code
-        loc = dict(func=func)
+        loc = dict(model=model)
         exec(code, loc, loc)
-        model = loc['model']
-
-        return model
-
-    def _calc_distances(self, force=False):
-        if self._dist is not None and not force:
-            return
-
-        # if self._X is of just one dimension, concat zeros.
-        if self._X.ndim == 1:
-            _x = np.vstack(zip(self._X, np.zeros(len(self._X))))
-        else:
-            _x = self._X
-        # else calculate the distances
-        self._dist = self._dist_func_wrapper(_x)
+        return loc['fitted_model']
 
     def _calc_diff(self, force=False):
         """Calculates the pairwise differences
@@ -1411,10 +1453,34 @@ class Variogram(object):
             return
 
         v = self.values
+        
+        # handle sparse matrix
+        if isinstance(self.distance_matrix, sparse.spmatrix):
+            c = r = self.triangular_distance_matrix
+            if not isinstance(c, sparse.csr.csr_matrix):
+                c = c.tocsr()
+            if not isinstance(r, sparse.csc.csc_matrix):
+                r = r.tocsc()
+            Vcol = sparse.csr_matrix(
+                (self.values[c.indices], c.indices, c.indptr)
+            )
+            Vrow = sparse.csc_matrix(
+                (self.values[r.indices], r.indices, r.indptr)
+            ).tocsr()
 
-        # Append a column of zeros to make pdist happy
-        # euclidean: sqrt((a-b)**2 + (0-0)**2) == sqrt((a-b)**2) == abs(a-b)
-        self._diff = pdist(np.column_stack((v, np.zeros(len(v)))), metric="euclidean")
+            # self._diff will have same shape as self.distances, even
+            # when that's not in diagonal format...
+            # Note: it might be compelling to do np.abs(Vrow -
+            # Vcol).data instead here, but that might optimize away
+            # some zeros, leaving _diff in a different shape
+            self._diff = np.abs(Vrow.data - Vcol.data)
+        else:
+            # Append a column of zeros to make pdist happy
+            # euclidean: sqrt((a-b)**2 + (0-0)**2) == sqrt((a-b)**2) == abs(a-b)
+            self._diff = pdist(
+                np.column_stack((v, np.zeros(len(v)))),
+                metric="euclidean"
+            )
 
     def _calc_groups(self, force=False):
         """Calculate the lag class mask array
@@ -1663,8 +1729,7 @@ class Variogram(object):
 
         # calculate the residuals
         return np.fromiter(
-            map(lambda x, y: x - y, model, experimental),
-            np.float
+            map(lambda x, y: x - y, model, experimental), float
         )
 
     @property
@@ -1678,7 +1743,7 @@ class Variogram(object):
         -------
         float
         """
-        return np.nanmean(np.fromiter(map(np.abs, self.residuals), np.float))
+        return np.nanmean(np.fromiter(map(np.abs, self.residuals), float))
 
     @property
     def rmse(self):
@@ -1709,9 +1774,8 @@ class Variogram(object):
 
         # get the sum of squares
         rsum = np.nansum(np.fromiter(
-            map(lambda x, y: (x - y)**2, experimental, model),
-            np.float
-        ))
+            map(lambda x, y: (x - y)**2, experimental, model), float)
+        )
 
         return np.sqrt(rsum / len(model))
 
@@ -1788,11 +1852,17 @@ class Variogram(object):
         mx = np.nanmean(experimental)
         my = np.nanmean(model)
 
-        # claculate the single pearson correlation terms
-        term1 = np.nansum(np.fromiter(map(lambda x, y: (x-mx) * (y-my), experimental, model), np.float))
+        # calculate the single pearson correlation terms
+        term1 = np.nansum(np.fromiter(
+            map(lambda x, y: (x-mx) * (y-my), experimental, model), float
+        ))
 
-        t2x = np.nansum(np.fromiter(map(lambda x: (x-mx)**2, experimental), np.float))
-        t2y = np.nansum(np.fromiter(map(lambda y: (y-my)**2, model), np.float))
+        t2x = np.nansum(
+            np.fromiter(map(lambda x: (x-mx)**2, experimental), float)
+        )
+        t2y = np.nansum(
+            np.fromiter(map(lambda y: (y-my)**2, model), float)
+        )
 
         return term1 / (np.sqrt(t2x * t2y))
 
@@ -1807,8 +1877,8 @@ class Variogram(object):
         mx = np.nanmean(experimental)
 
         # calculate the single nash-sutcliffe terms
-        term1 = np.nansum(np.fromiter(map(lambda x, y: (x - y)**2, experimental, model), np.float))
-        term2 = np.nansum(np.fromiter(map(lambda x: (x - mx)**2, experimental), np.float))
+        term1 = np.nansum(np.fromiter(map(lambda x, y: (x - y)**2, experimental, model), float))
+        term2 = np.nansum(np.fromiter(map(lambda x: (x - mx)**2, experimental), float))
 
         return 1 - (term1 / term2)
 
@@ -1848,26 +1918,26 @@ class Variogram(object):
         Return a dictionary of the variogram parameters.
 
         .. versionchanged:: 0.3.7
-            The describe now returns all init parameters in as the 
+            The describe now returns all init parameters in as the
             `describe()['params']` key and all keyword arguments as
-            `describe()['kwargs']`. This output can be suppressed 
+            `describe()['kwargs']`. This output can be suppressed
             by setting `short=True`.
-        
+
         Parameters
         ----------
         short : bool
-            If `True`, the `'params'` and `'kwargs'` keys will be 
+            If `True`, the `'params'` and `'kwargs'` keys will be
             omitted. Defaults to `False`.
         flat : bool
             If `True`, the `'params'` and `'kwargs'` nested `dict`s
-            will be distributed to the main `dict` to return a 
+            will be distributed to the main `dict` to return a
             flat `dict`. Defaults to `False`
 
         Returns
         -------
         parameters : dict
             Returns fitting parameters of the theoretical variogram
-            model along with the init parameters of the 
+            model along with the init parameters of the
             `Variogram <skgstat.Variogram>` instance.
 
         """
@@ -1876,23 +1946,35 @@ class Variogram(object):
             self.fit(force=True)
 
         # scale sill and range
-        if self.normalized:
-            maxlag = np.nanmax(self.bins)
-            maxvar = np.nanmax(self.experimental)
-        else:
-            maxlag = 1.
-            maxvar = 1.
+        maxlag = np.nanmax(self.bins)
+        maxvar = np.nanmax(self.experimental)
 
         # get the fitting coefficents
         cof = self.cof
 
         # build the dict
+
+        def fnname(fn):
+            if callable(fn):
+                name = "%s.%s" % (fn.__module__, fn.__name__)
+            else:
+                name = fn
+            if name.startswith("skgstat."):
+                return name.split(".")[-1]
+            return name
+
         rdict = dict(
-            name=self._model.__name__,
-            estimator=self._estimator.__name__,
-            effective_range=cof[0] * maxlag,
-            sill=cof[1] * maxvar,
-            nugget=cof[-1] * maxvar if self.use_nugget else 0
+            model=fnname(self._model) if not self._harmonize else "harmonize",
+            estimator=fnname(self._estimator),
+            dist_func=fnname(self.dist_function),
+
+            normalized_effective_range=cof[0] * maxlag,
+            normalized_sill=cof[1] * maxvar,
+            normalized_nugget=cof[-1] * maxvar if self.use_nugget else 0,
+
+            effective_range=cof[0],
+            sill=cof[1],
+            nugget=cof[-1] if self.use_nugget else 0,
         )
 
         # handle s parameters for matern and stable model
@@ -1907,7 +1989,7 @@ class Variogram(object):
             params = dict(
                 estimator=self._estimator.__name__,
                 model=self._model.__name__,
-                dist_func=str(self._dist_func_name),
+                dist_func=str(self.dist_function),
                 bin_func=self._bin_func_name,
                 normalize=self.normalized,
                 fit_method=self.fit_method,
@@ -1995,6 +2077,36 @@ class Variogram(object):
             self._model.__name__: data}
         ).copy()
 
+    def to_gstools(self, **kwargs):
+        """
+        Instantiate a corresponding GSTools CovModel.
+
+        By default, this will be an isotropic model.
+
+        Parameters
+        ----------
+        **kwargs
+            Keyword arguments forwarded to the instantiated GSTools CovModel.
+            The default parameters 'dim', 'var', 'len_scale', 'nugget',
+            'rescale' and optional shape parameters will be extracted
+            from the given Variogram but they can be overwritten here.
+
+        Raises
+        ------
+        ImportError
+            When GSTools is not installed.
+        ValueError
+            When GSTools version is not v1.3 or greater.
+        ValueError
+            When given Variogram model is not supported ('harmonize').
+
+        Returns
+        -------
+        :any:`CovModel`
+            Corresponding GSTools covmodel.
+        """
+        return skgstat_to_gstools(self, **kwargs)
+
     def plot(self, axes=None, grid=True, show=True, hist=True):
         """Variogram Plot
 
@@ -2057,14 +2169,14 @@ class Variogram(object):
         Parameters
         ----------
         ax : matplotlib.Axes, plotly.graph_objects.Figure
-            If None, a new plotting Figure will be created. If given, 
-            it has to be an instance of the used plotting backend, which 
+            If None, a new plotting Figure will be created. If given,
+            it has to be an instance of the used plotting backend, which
             will be used to plot on.
         show : boolean
-            If True (default), the `show` method of the Figure will be 
-            called. Can be set to False to prevent duplicated plots in 
+            If True (default), the `show` method of the Figure will be
+            called. Can be set to False to prevent duplicated plots in
             some environments.
-        
+
         Returns
         -------
         fig : matplotlib.Figure, plotly.graph_objects.Figure
@@ -2118,12 +2230,12 @@ class Variogram(object):
 
             .. note::
                 Right now, this is only supported for ``'plotly'`` backend
-   
+
 
         Returns
         -------
         fig : matplotlib.Figure, plotly.graph_objects.Figure
-            The figure produced by the function. Dependends on the 
+            The figure produced by the function. Dependends on the
             current backend.
 
         """
@@ -2134,7 +2246,7 @@ class Variogram(object):
             return plotting.matplotlib_location_trend(self, axes=axes, show=show, **kwargs)
         elif used_backend == 'plotly':
             return plotting.plotly_location_trend(self, fig=axes, show=show, **kwargs)
-        
+
         # if we reach this line, somethings wrong with plotting backend
         raise ValueError('The plotting backend has an undefined state.')
 
@@ -2207,7 +2319,7 @@ class Variogram(object):
         _range = np.NaN if 'error' in par else par['effective_range']
         _nugget = np.NaN if 'error' in par else par['nugget']
 
-        s = "{0} Variogram\n".format(par['name'])
+        s = "{0} Variogram\n".format(par['model'])
         s+= "-" * (len(s) - 1) + "\n"
         s+="""Estimator:         %s
         \rEffective Range:   %.2f
