@@ -15,10 +15,51 @@ except ImportError:
     PLOTLY_FOUND = False
 
 from skgstat import Variogram
+from skgstat import OrdinaryKriging
 from skgstat import estimators
+from skgstat import models
 from skgstat import plotting
 
 
+class TestSpatiallyCorrelatedData(unittest.TestCase):
+    def setUp(self):
+        # Generate some random but spatially correlated data
+        # with a range of ~20
+        
+        np.random.seed(42)
+        c = np.random.sample((50, 2)) * 60
+        np.random.seed(42)
+        v = np.random.normal(10, 4, 50)
+        
+        V = Variogram(c, v).describe()
+        V["effective_range"] = 20
+        OK = OrdinaryKriging(V, coordinates=c, values=v)
+
+        self.c = np.random.sample((500, 2)) * 60
+        self.v = OK.transform(self.c)
+
+        self.c = self.c[~np.isnan(self.v),:]
+        self.v = self.v[~np.isnan(self.v)]
+
+    def test_dense_maxlag_inf(self):
+        Vdense = Variogram(self.c, self.v)
+        Vsparse = Variogram(self.c, self.v, maxlag=10000000)
+
+        for x, y in zip(Vdense.parameters, Vsparse.parameters):
+            self.assertAlmostEqual(x, y, places=3)
+            
+    def test_sparse_maxlag_50(self):
+        V = Variogram(self.c, self.v, maxlag=50)
+
+        for x, y in zip(V.parameters, [20.264, 6.478, 0]):
+            self.assertAlmostEqual(x, y, places=3)
+            
+    def test_sparse_maxlag_30(self):
+        V = Variogram(self.c, self.v, maxlag=30)
+
+        for x, y in zip(V.parameters, [17.128, 6.068, 0]):
+            self.assertAlmostEqual(x, y, places=3)
+
 class TestVariogramInstatiation(unittest.TestCase):
     def setUp(self):
         # set up default values, whenever c and v are not important
@@ -32,6 +73,25 @@ class TestVariogramInstatiation(unittest.TestCase):
 
         for x, y in zip(V.parameters, [7.122, 13.966, 0]):
             self.assertAlmostEqual(x, y, places=3)
+            
+    def test_sparse_standard_settings(self):
+        V = Variogram(self.c, self.v, maxlag=10000)
+
+        for x, y in zip(V.parameters, [7.122, 13.966, 0]):
+            self.assertAlmostEqual(x, y, places=3)
+
+    def test_input_dimensionality(self):
+        c1d = np.random.normal(0, 1, 100)
+        c3d = np.random.normal(0, 1, size=(100, 3))
+        v = np.random.normal(10, 4, 100)
+
+        # test 1D coords
+        V = Variogram(c1d, v)
+        self.assertTrue(V.dim == 1)
+
+        # test 3D coords
+        V2 = Variogram(c3d, v)
+        self.assertTrue(V2.dim == 3)
 
     def test_pass_median_maxlag_on_instantiation(self):
         np.random.seed(1312)
@@ -63,6 +123,13 @@ class TestVariogramInstatiation(unittest.TestCase):
             str(e.exception)
         )
 
+    def test_invalid_binning_func(self):
+        with self.assertRaises(AttributeError) as e:
+            V = Variogram(self.c, self.v)
+            V.set_bin_func(42)
+        
+        self.assertTrue('of type string' in str(e.exception))
+
     def test_unknown_model(self):
         with self.assertRaises(ValueError) as e:
             Variogram(self.c, self.v, model='unknown')
@@ -289,18 +356,22 @@ class TestVariogramArguments(unittest.TestCase):
     def test_callable_dist_function(self):
         V = Variogram([(0, 0), (4, 1), (1, 1)], [1, 2, 3], n_lags=2)
 
-        def dfunc(x):
-            return np.ones((len(x), len(x)))
+        def dfunc(u, v):
+            return 1
 
         V.set_dist_function(dfunc)
 
         # test
         self.assertEqual(V.dist_function, dfunc)
         self.assertTrue((V.distance==1).all())
-        self.assertEqual(V.distance.shape, (3, 3))
+        self.assertEqual(V.distance_matrix.shape, (3, 3))
 
     @staticmethod
-    def test_direct_dist_setting():
+    def disabled_test_direct_dist_setting():
+        # Distance can no longer be explicitly set
+        # it would require setting the whole MetricSpace, with a
+        # non-sparse diagonal matrix
+        
         V = Variogram([(0, 0), (4, 1), (1, 1)], [1, 2, 3], n_lags=2)
 
         V.distance = np.array([0, 0, 100])
@@ -332,7 +403,7 @@ class TestVariogramArguments(unittest.TestCase):
         V.use_nugget = True
         self.assertEqual(V.use_nugget, True)
         self.assertEqual(V._use_nugget, True)
-        self.assertAlmostEqual(V.describe()['nugget'], 291.28, places=2)
+        self.assertAlmostEqual(V.describe()['normalized_nugget'], 291.28, places=2)
 
     def test_use_nugget_exception(self):
         with self.assertRaises(ValueError) as e:
@@ -728,8 +799,8 @@ class TestVariogramQaulityMeasures(unittest.TestCase):
         V = Variogram(self.c, self.v)
 
         for model, rmse in zip(
-                ['spherical', 'gaussian', 'matern', 'stable'],
-                [3.3705, 3.3707, 3.1811, 3.193]
+                ['spherical', 'gaussian', 'stable'],
+                [3.3705, 3.3707, 3.193]
         ):
             V.set_model(model)
             self.assertAlmostEqual(V.rmse, rmse, places=4)
@@ -738,8 +809,8 @@ class TestVariogramQaulityMeasures(unittest.TestCase):
         V = Variogram(self.c, self.v)
 
         for model, mr in zip(
-            ['spherical', 'cubic', 'matern', 'stable'],
-            [2.6803, 2.6803, 2.6701, 2.6966]
+            ['spherical', 'cubic', 'stable'],
+            [2.6803, 2.6803, 2.6966]
         ):
             V.set_model(model)
             self.assertAlmostEqual(V.mean_residual, mr, places=4)
@@ -858,8 +929,12 @@ class TestVariogramMethods(unittest.TestCase):
             [0., 11.82, 13.97, 13.97, 13.97, 13.97, 13.97, 13.97, 13.97, 13.97],
             decimal=2
         )
-    
-    def test_data_with_force(self):
+
+    def disabled_test_data_with_force(self):
+        # Distance can no longer be explicitly set
+        # it would require setting the whole MetricSpace, with a
+        # non-sparse diagonal matrix
+        
         # should work if _dist is corccupted
         self.V._dist = self.V._dist * 5.
         self.V.cof = None
@@ -900,7 +975,7 @@ class TestVariogramMethods(unittest.TestCase):
         V = self.V.clone()
         
         # test matern
-        param = [42.3, 15.42, 0.11, 0.]
+        param = [42.3, 16.2, 0.1, 0.]
         V.set_model('matern')
         assert_array_almost_equal(V.parameters, param, decimal=2)
     
@@ -908,7 +983,7 @@ class TestVariogramMethods(unittest.TestCase):
         V = self.V.clone()
 
         # test stable
-        param = [42.3 , 15.79, 0.45,  0.]
+        param = [42.3, 15.79, 0.45,  0.]
         V.set_model('stable')
         assert_array_almost_equal(V.parameters, param, decimal=2)
 
